###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import json
import logging
import os.path

import click

from csv2bufr import __version__, BUFRMessage, transform as transform_csv

THISDIR = os.path.dirname(os.path.realpath(__file__))
MAPPINGS = f"{THISDIR}{os.sep}resources{os.sep}mappings"

# configure logging
LOGGER = logging.getLogger()
LOGGER.setLevel(logging.DEBUG)


def cli_option_verbosity(f):
    logging_options = ["ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]

    def callback(ctx, param, value):
        if value is not None:
            LOGGER.setLevel(getattr(logging, value))
        return True

    return click.option("--verbosity", "-v",
                        type=click.Choice(logging_options),
                        help="Verbosity",
                        callback=callback)(f)


def cli_callbacks(f):
    f = cli_option_verbosity(f)
    return f


@click.group()
@click.version_option(version=__version__)
def cli():
    """csv2bufr"""
    pass


@click.group()
def data():
    """data workflows"""
    pass


@click.group()
def mappings():
    """stored mappings"""
    pass


@click.command('list')
@click.pass_context
def list_mappings(ctx):
    for mapping in os.listdir(MAPPINGS):
        msg = f"{mapping} => {MAPPINGS}{os.sep}{mapping}"
        click.echo(msg)


@click.command('create')
@click.pass_context
@click.argument("sequence", nargs=-1, type=int)
@click.option("--output", "output", help="File to save the template to")
@cli_option_verbosity
def create_mappings(ctx, sequence, output, verbosity):
    msg = BUFRMessage(sequence)
    template = msg.create_template()
    if output:
        with open(output, "w") as fh:
            fh.write(json.dumps(template, indent=4))
    else:
        print(json.dumps(template, indent=4))


@click.command()
@click.pass_context
@click.argument("csv_file", type=click.File(errors="ignore"))
@click.option("--bufr-template", "mapping", required=True,
              help="Name of file or mapping template to use to map from CSV to BUFR")  # noqa
@click.option("--output-dir", "output_dir", required=False,
              help="Name of output file", default='.')
@cli_option_verbosity
def transform(ctx, csv_file, mapping, output_dir, verbosity):  # noqa
    result = None
    click.echo(f"\nCLI:\t... Transforming {csv_file.name} to BUFR ...")

    # identify mapping to use
    if not os.path.isfile(mapping):
        mappings_file = f"{MAPPINGS}{os.sep}{mapping}.json"
        if not os.path.isfile(mappings_file):
            raise click.ClickException(
                f"Invalid stored mapping ({mappings_file})")
    else:
        mappings_file = mapping

    # load mappings
    with open(mappings_file) as fh:
        mappings = json.load(fh)

    try:
        result = transform_csv(csv_file.read(), mappings)
    except Exception as err:
        raise click.ClickException(err)

    click.echo("CLI:\t... Processing subsets:")
    for item in result:
        key = item['_meta']["id"]
        bufr_filename = f"{output_dir}{os.sep}{key}.bufr4"
        if item['bufr4'] is not None:
            try:
                with open(bufr_filename, "wb") as fh:
                    fh.write(item["bufr4"])
                    nbytes = fh.tell()

                click.echo(f"CLI:\t..... {nbytes} bytes written to {bufr_filename}")  # noqa
            except Exception as err:
                raise click.ClickException(err)
        else:
            click.echo("CLI:\t..... 'None' found in BUFR output, no data written")  # noqa

    click.echo("CLI:\tEnd of processing, exiting.\n")


data.add_command(transform)
mappings.add_command(list_mappings)
mappings.add_command(create_mappings)

cli.add_command(data)
cli.add_command(mappings)



#
# Name: xmlConfig.py
# Purpose: To provide TAC->IWXXM and TAC->USWX decoders & encoders with
# site-specific information and customized settings
#
# Author: Mark Oberfield
# Organization: NOAA/NWS/OSTI/Meteorological Development Laboratory
# Contact Info: Mark.Oberfield@noaa.gov
# Date: 16 January 2020
#
import os
import re
#
os.environ['TZ'] = 'GMT0'
# -----------------------------------------------------------------------------------
#
# If this centre or office is performing bulk TAC->XML translations, or TAC->XML
# translations on behalf of another country, then TRANSLATOR should be set to true
# and its name and designator provided.
#
# For Meteorological or Space Weather Watch offices running this software, TRANSLATOR
# should be set to False.
#
TRANSLATOR = True
#
# The full name and identifier of translation centre running this software
#
TranslationCentreName = 'NCEP Central Operations'
TranslationCentreDesignator = 'KWNO'
#
# -----------------------------------------------------------------------------------
#
# The remainder of this file is relatively static, requiring updates when either:
# IWXXM versioning changes, WMO Code Registry and/or Annex 3 or WMO 306 changes.
#
# -----------------------------------------------------------------------------------
#
# IWXXM versioning
_iwxxm = '3.0'
_release = '3.0'
#
IWXXM_URI = 'http://icao.int/iwxxm/%s' % _iwxxm
IWXXM_URL = 'http://schemas.wmo.int/iwxxm/%s/iwxxm.xsd' % _release
#
# Path to file containing codes obtained from WMO Code Registry in RDF/XML format.
#
CodesFilePath = os.path.join(os.path.dirname(__file__), '../data')
#
# To support Annex 3 code forms, the following Containers from the WMO Code Registry
# site were downloaded into the CodesFilePath directory in RDF format.
#
# These are needed for MDL's IWXXM Encoders
#
CLDAMTS = 'CloudAmountReportedAtAerodrome'
COLOUR_CODES = 'AviationColourCode'
CVCTNCLDS = 'SigConvectiveCloudType'
NIL = 'nil'
RECENTWX = 'AerodromeRecentWeather'
RWYDEPST = '0-20-086'
RWYCNTMS = '0-20-087'
RWYFRCTN = '0-20-089'
SEACNDS = '0-22-061'
SWX_PHENOMENA = 'SpaceWxPhenomena'
SWX_LOCATION = 'SpaceWxLocation'
WEATHER = 'AerodromePresentOrForecastWeather'
#
# NIL reason values
MSSG = 'missing'
NA = 'inapplicable'
NOAUTODEC = 'notDetectedByAutoSystem'
NOOBSV = 'notObservable'
NOOPRSIG = 'nothingOfOperationalSignificance'
NOSIGC = 'noSignificantChange'
UNKNWN = 'unknown'
WTHLD = 'withheld'
#
# Coordinate Reference System used for all IWXXM messages. axisLabels and srsDimension
# must be consistent with CRS specified with srsName.
#
srsName = 'http://www.opengis.net/def/crs/EPSG/0/4326'
axisLabels = 'Lat Long'
srsDimension = '2'
#
# METAR/SPECI and TAF variables
# ---------------------------------------------------------------------------------------------
# If elevation of the aerodrome is provided, then provide the vertical datum it is
# based on and its unit of measure
#
useElevation = False
#
# Vertical datum must be set correctly for the elevation used. Allowed values are:
# 'EGM_96', 'AHD', 'NAVD88' or string matching the regular expression pattern:
#
# 'OTHER:(\w|_){1,58}'
#
verticalDatum = 'EGM_96'
#
# Elevation value unit of measure (UOM). Either 'FT' or 'M' or string matching the
# regular expression pattern:
#
# 'OTHER:(\w|_){1,58}'
#
elevationUOM = 'M'
#
# xlink:title attributes are optional in IWXXM XML documents. TITLES variable below
# determines whether they are displayed in TAF, METAR and SPECI reports
#
# If no xlink:title attributes are wanted in IWXXM XML documents, set TITLES to 0 (zero).
# Otherwise, set bits appropriately.
#
# Bit masks
Weather = 1 << 0
CloudAmt = 1 << 1
CloudType = 1 << 2
SeaCondition = 1 << 3
#
TITLES = 0
# TITLES=(CloudAmt|CloudType|SeaCondition)
#
# If xlink:titles are to appear in the document, set preferred language. English, 'en',
# is the default if the desired language is not found in the WMO Code Registry.
#
PreferredLanguageForTitles = 'en'
#
# Variables affecting METAR/SPECI Runway conditions. Depreciated, but keep for now.
#
NIL_SNOCLO_URL = 'http://codes.wmo.int/bufr4/codeflag/0-20-085/1'
RunwayDeposit = 1 << 4
AffectedRunwayCoverage = 1 << 5
RunwayFriction = 1 << 6
RunwayFrictionValues = {'91': 'Braking action poor', '92': 'Braking action medium to poor',
                        '93': 'Braking action medium', '94': 'Braking action medium to good',
                        '95': 'Braking action good', '99': 'Unreliable'}
#
# If prevailing horizontal visibility falls below this value (metres), RVR information
# should be supplied
#
RVR_MaximumDistance = 1500
#
# Critera for including sector visibility under the iwxxm namespace. See Recommendation
# 4.2.4.4a in Annex 3
#
Max_SectorVisibility_1 = 1500
Max_SectorVisibility_2 = 5000
Max_PercentageOfPrevailing = 50
#
# The use of CAVOK introduces potential ambiguities when conditions are forecasted to
# change during the TAF period of validity.
#
# When a forecast group introduces changes to sky or visibility conditions (and not both at
# the same time), it is not clear how the other elment is affected and the IWXXM schematron
# rule, TAF.TAF-8, requires both to be known when CAVOK is not in effect.
#
# By setting the 'noImpliedCAVOKCondition' to True, the TAF decoder will flag this forecast
# group to the forecaster for remediation.
#
# If noImpliedCAVOKCondition is set to False, then the missing element will be inserted
# into the forecast.  If visibility is missing, it will be set to >10km.  If sky condition
# is missing, it will be set to 'NSC' (no significant cloud).  By default, a message
# will be written to the log file indicating this action. By setting
# emitImpliedCAVOKConditionMessages, these messages will not be written to the log file.
#
noImpliedCAVOKCondition = True
emitImpliedCAVOKConditionMessage = True  # Only effective when noImpliedCAVOKCondition is False
# ---------------------------------------------------------------------------------------------
# Information is needed for Space Weather Advisories
#
# DAYLIGHTSIDE is the radius of the circle that circumscribes the daylight side of the
# Earth which is approximately one-quarter of the Earth's circumference
#
DAYLIGHTSIDE = 'DAYLIGHT_SIDE'
DAYLIGHTSIDE_RADIUS = '10100'
#
# DAYLIGHTSIDE_UOM, only [mi_i] or 'km' is allowed
DAYLIGHTSIDE_UOM = 'km'
#
# Whether latitude bands in SWX product are combined
JOIN_BANDS = False
# -----------------------------------------------------------------------------------
#
# IWXXM-US versioning
_us_iwxxm = '3.0'
_us_iwxxm_release = '3.0'
#
# USWX versioning
_uswx = '1.0'
_uswx_release = '1.0'
#
# US Code Registry for Meteorological Services
OFCM_CODE_REGISTRY_URL = 'https://codes.nws.noaa.gov'
#
# IWXXM-US URI and URLs
IWXXM_US_URI = 'http://www.weather.gov/iwxxm-us/%s' % _us_iwxxm
IWXXM_US_URL = 'https://nws.weather.gov/schemas/iwxxm-us/%s/' % _us_iwxxm_release
IWXXM_US_METAR_URL = 'https://nws.weather.gov/schemas/iwxxm-us/%s/metarSpeci.xsd' % _us_iwxxm_release
#
# USWX_URI and URLs
USWX_URI = 'http://nws.weather.gov/schemas/USWX/%s' % _uswx
USWX_URL = 'https://nws.weather.gov/schemas/uswx/%s' % _uswx_release
#
# For non-Annex 3 content in US METAR/SPECI reports
ACCUMLTN = 'http://codes.wmo.int/grib2/codeflag/4.10/1'
MAXIMUM = 'http://codes.wmo.int/grib2/codeflag/4.10/2'
MINIMUM = 'http://codes.wmo.int/grib2/codeflag/4.10/3'
AERONAUTICALVIS = 'http://codes.wmo.int/common/quantity-kind/aeronauticalVisibility'

PRSRTNDCY = '0-10-063'
SKYCATALOG = '0-20-012'
#
# Surface observations with 'ICAO' identifers that match the following regular expression are considered
# to be based on FMH-1 or AFMAN 15-111 manuals. These patterns should only match aerodrome observations that
# are issued from the United States or overseas military bases.
#
CONUS = 'K[A-Z0-9]{3}'
ALASKA = 'P(A|F|O|P)[A-Z]{2}'
PACIFIC = 'P(G|H|K|L|M|W)[A-Z]{2}'
SAN_JUAN = 'TJ[A-Z]{2}|TIST|TISX|TKPK|TNCM'
USMIL_EURO = 'BGTL|EGUL|EGUN|EGVA|ETAD|ETAR'
USMIL_CUBA = 'MUGM'
USMIL_ME = 'OEPS|OTBH'
USMIL_ASIA = 'RJTY|RKJK|RKSG|RKSO|RKSY|RODN'
#
# Combine all identifiers into a list
FMH1_AFMAN = [CONUS, ALASKA, PACIFIC, SAN_JUAN, USMIL_EURO, USMIL_ASIA, USMIL_CUBA, USMIL_ME]
#
# Create the regular expression
isUSObservation = re.compile(r'(%s)' % '|'.join(FMH1_AFMAN))



import re

import metar2bufr.metarDecoder.metarDecoders as MD
import metar2bufr.metarDecoder.xmlUtilities as deu

from copy import deepcopy
from typing import Iterator
import csv
from io import StringIO
import logging
import json
import math
import os
from csv2bufr import BUFRMessage
import urllib.request as request
# from FM15_bfr2geojson import transform as as_geojson
from FM15_bfr2geojson import generic_bufr2geojson
LOGGER = logging.getLogger(__name__)

FAILED = 0
PASSED = 1

fmh = MD.FMH1()
annex3 = MD.Annex3()

re_ID = re.compile(r'(METAR|SPECI)\s+(COR\s+)?(?P<id>\w{4})')

# metar = "METAR NZFX 032155Z 00000KT 9999 BCFG FEW010 FEW030 M26/M29 A2873 RMK VIS GRID N 1600 SLP724 LSR10 SDN/HDP HDN GRID N-E="
# metar = "METAR KRZT 221615Z AUTO 08007KT 10SM SCT023 OVC038 20/13 A3003 RMK AO2 PWINO="
# metar = "METAR SPTN 102000Z 23013KT 9999 SCT050 SCT120 27/17 Q1013 RMK PP000="
# metar = "METAR KOGS 221610Z AUTO VRB03G11KT 10SM FEW049 26/14 A3017 RMK AO2="
# metar = "METAR KOWX 221615Z AUTO 08010KT 10SM BKN033 23/11 A3006 RMK AO1="
# metar = "METAR KABC 121755Z AUTO 21016G24KT 180V240 1SM R11/P6000FT -RA BR BKN015 OVC025 06/04 A2990="
# BKN015 OVC025
# BKN015TCU OVC025CB

metar = "METAR KAFF 121755Z COR 210P16G24KT 180V240 1SM R11/6000FTU R12/3000D R01L/M0600V6000FT -RA VA PO VV027 06/04 A2990="
# metar = "METAR KABC 121755Z AUTO 21016G24KT 180V240 1SM R11/P6000FT -RA BR BKN015 OVC025 06/04 A2990="

# metar = "METAR KAFF 121755Z COR 210P16G24KT 180V240 1000 R11/6000FT R12/P6000 R01L/0600VP6000FT -RA VA PO SKC 06/04 A2990 RMK VIS 3/4V1 1/2="



_keys = ['tsi', 'product_status', 'station_type',
         'year', 'month', 'day', 'hour', 'minute',
         'wind_direction', 'variable_extreme_ccw', 'variable_extreme_cw',
         'wind_speed_qualifier', 'wind_speed_kn', 'wind_speed_m/s'
         'wind_gust_qualifier', 'gust_speed_kn', 'gust_speed_m/s'
         'air_temp', 'dew_point_temp', 'altimeter', 'general_weather_indicator', 'prevailing_visibility']

metar_template = dict.fromkeys(_keys)

THISDIR = os.path.dirname(os.path.realpath(__file__))
MAPPINGS = f"{THISDIR}{os.sep}resources{os.sep}307051_metar_mapping.json"
ICAOSTATIONSURL = "https://www.aviationweather.gov/docs/metar/stations.txt"

# MAPPINGS = "/home/alexander.thompson/metar2bufr/src/metar-mappings2.json"

# Load template mappings file, this will be updated for each message.
with open(MAPPINGS) as fh:
    _mapping = json.load(fh)


def update_data_mapping(mapping: list, update: dict):
    match = False
    for idx in range(len(mapping)):
        if mapping[idx]['eccodes_key'] == update['eccodes_key']:
            match = True
            break
    if match:
        mapping[idx] = update
    else:
        mapping.append(update)
    return mapping

def extract_metar(data: str) -> list:
    if not data.__contains__("="):
        regex = re.compile(r'MTR.{3}')
        if regex.search(data):
            pass
        else:
            LOGGER.error((
                "Delimiters (=) are not present in the string"))
            LOGGER.debug(data)
            raise ValueError
    
    start_position = data.find("METAR")
    if start_position == -1:
        start_position = data.find("SPECI")
        if start_position == -1:
            raise ValueError("Invalid METAR message. 'METAR' or 'SPECI' could not be found.")
        
    data = re.split('=', data[start_position:])

    for i in range(len(data)):
        data[i] = data[i]+"="

    # print(data[:len(data)-1])
    return data[:len(data)-1]

def parse_metar(message: str, year: int, month: int) -> dict:
    message = message.strip()
    LOGGER.debug(f"Parsing message: {message}")
    icaoID = re_ID.match(message).group('id')
    if deu.isUSIdentifier(icaoID):
        decoded = fmh(message)
    else:
        decoded = annex3(message)

    for key in decoded:
        LOGGER.debug(key, ': ', decoded[key])

    output = deepcopy(metar_template)

    # output['report_type'] = message[0:5]
    # output['year'] = year
    # output['month'] = month

    #   ecCodes 001063: ICAO Location Indicator
    if 'ident' in decoded:
        icao = decoded['ident']['str']
        output['icao_location_identifier'] = icao

        try: 
            output['tsi'] = icao
            # output['block_no'] = tsi[0:2]
            # output['station_no'] = tsi[2:5]
        except Exception:
            # tsi = None
            output['tsi'] = None
            # output['block_no'] = None
            # output['station_no'] = None

    #   ecCodes 008079: Aviation Product Status
    #   first get report type which will either be normal (METAR) or special (SPECI)
    report_type = message[0:5]
    #   codes for normal/routine reports
    if report_type == 'METAR':
        #   correction to previously reported product, coded as 1
        if 'cor' in decoded:
            output['product_status'] = 1
        #   normal issue coded as 0
        else:
            output['product_status'] = 0
    #   codes for special reports
    elif report_type == 'SPECI':
        #   corrected special report
        if 'cor' in decoded:
            output['product_status'] = 7
        #   original special report
        else:
            output['product_status'] = 6
    #   TODO: figure out NIL coding

    #   ecCodes 002001: Station Type
    #   metar station type is binary, either AUTOMATIC (0) or MANNED (1)
    if 'auto' in decoded:
        output['station_type'] = 0
    else:
        output['station_type'] = 1

    #   ecCodes 301011: year, month, day / 301012: hour, minute
    output['year'] = year
    output['month'] = month
    if 'itime' in decoded:
        output['day'] = decoded['itime']['tuple'].tm_mday
        output['hour'] = decoded['itime']['tuple'].tm_hour
        output['minute'] = decoded['itime']['tuple'].tm_min

    if 'wind' in decoded:
        #   ecCodes 011001: wind direction
        #   check for missing value
        if decoded['wind']['dd'] != '///':
            output['wind_direction'] = decoded['wind']['dd']

        #   ecCodes 011016: Extreme counterclockwise wind direction of a variable wind
        if 'ccw' in decoded['wind']:
            output['variable_extreme_ccw'] = int(decoded['wind']['ccw'])

        #   ecCodes 011017: Extreme clockwise wind direction of a variable wind
        if 'cw' in decoded['wind']:
            output['variable_extreme_cw'] = int(decoded['wind']['cw'])

        #   ecCodes 008054: qualifier for wind speed
        #   wind speed is greater than what is reported
        if 'ffplus' in decoded['wind']:
            output['wind_speed_qualifier'] = 1
        #   wind speed is as reported
        else:
            output['wind_speed_qualifier'] = 0

        wind_uom = decoded['wind']['uom'] 
        wind_speed = decoded['wind']['ff']

        #   check for missing value
        if wind_speed != '//':
            #   ecCodes 011084: wind speed in knots
            if wind_uom == '[kn_i]':
                output['wind_speed_kn'] = int(wind_speed)

            #   ecCodes 011002: wind speed in m/s
            elif wind_uom == 'm/s':
                output['wind_speed_m/s'] = int(wind_speed)

        #   ecCodes 008054: qualifier for wind gusts
        #   wind gusts are greater than what is reported
        if 'ggplus' in decoded['wind']:
            output['wind_gust_qualifier'] = 1
        #   wind gusts are as reported
        else:
            output['wind_gust_qualifier'] = 0

        if 'gg' in decoded['wind']:
            gust_speed = decoded['wind']['gg']
            #   ecCodes 011086: Maximum wind gust speed reported in knots
            if wind_uom == '[kn_i]':
                output['gust_speed_kn'] = int(gust_speed)
            #   ecCodes 011041: Maximum wind gust speed reported in m/s
            elif wind_uom == 'm/s':
                output['gust_speed_m/s'] = int(gust_speed)

    if 'temps' in decoded:
        if 'air' in decoded['temps']:
            #   check for missing value
            if decoded['temps']['air'] != '//':
                #   ecCodes 012023: Air Temperature in Celsius
                output['air_temp'] = float(decoded['temps']['air'])
        if 'dewpoint' in decoded['temps']:
            #   check for missing value
            if decoded['temps']['dewpoint'] != '//':
                #   ecCodes 012024: Dewpoint temperature in Celsius
                output['dew_point_temp'] = float(decoded['temps']['dewpoint'])

    #   ecCodes 010052: Altimeter setting qnh in Pascals
    if 'altimeter' in decoded:
        altimeter = float(decoded['altimeter']['value'])
        alt_uom = decoded['altimeter']['uom']
        if alt_uom == 'hPa':
            output['altimeter'] = altimeter*100
        elif alt_uom == "[in_i'Hg]":
            # if metar records altimeter in inches of Mercury. need to convert to Pascals for bufr
            output['altimeter'] = round(altimeter*3386.39, 2)

    #   ecCodes 020009: General weather indicator of TAF or METAR
    if 'cavok' in decoded:
        output['general_weather_indicator'] = 2

    #   ecCodes 020060: Prevailing horizontal visibility in meters
    if 'vsby' in decoded:
        #   check for missing value
        if decoded['vsby']['value'] != '////':
            # if recorded in number of statute miles, need to convert value to meters to be bufr compatible
            if decoded['vsby']['uom'] == '[mi_i]':
                vis_miles = float(decoded['vsby']['value'])
                output['prevailing_visibility'] = math.floor(vis_miles * 1609.34)
            elif decoded['vsby']['uom'] == 'm':
                output['prevailing_visibility'] = int(decoded['vsby']['value'])

    #   ecCodes 020059: Minimum horizontal visibility in meters
    if 'vvis' in decoded:
        low_vis = float(decoded['vvis']['lo'])
        high_vis = float(decoded['vvis']['hi'])
        #   metar reports variable visibility in statute miles. need to convert to bufr
        output['low_vis'] = math.floor(low_vis * 1609.34)
        output['high_vis'] = math.floor(high_vis * 1609.34)

    num_rvr = 0
    if 'rvr' in decoded:
        num_rvr = len(decoded['rvr']['str'])
        for i in range(num_rvr):
            #   ecCodes 001064: Runway designator
            output[f'rvr_{i+1}_designator'] = decoded['rvr']['rwy'][i]
            qualifier = decoded['rvr']['oper'][i]
            #   ecCodes 008014: Qualifier for runway visual range
            #   encoding: 0 - NORMAL, 1 - ABOVE THE UPPER LIMIT, 2 - BELOW THE LOWER LIMIT
            if qualifier == 'ABOVE':
                output[f'rvr_{i+1}_qualifier'] = 1
            elif qualifier == 'BELOW':
                output[f'rvr_{i+1}_qualifier'] = 2
            elif qualifier == None:
                output[f'rvr_{i+1}_qualifier'] = 0

            #   ecCodes 020061: Runway Visual Range recorded in meters
            mean_vis = float(decoded['rvr']['mean'][i])
            # if distance is reported in meters, we can copy directly to bufr
            if decoded['rvr']['uom'][i] == 'm':
                output[f'rvr_{i+1}_mean'] = int(mean_vis)
            # if distance is reported in ft, have to convert to meters for bufr
            elif decoded['rvr']['uom'][i] == '[ft_i]':
                output[f'rvr_{i+1}_mean'] = math.floor(mean_vis*0.3048)

            #   ecCodes 020018: Tendency of runway visual range
            #   encoding: INCREASING -> 0, DECREASING -> 1, NO DISTINCT CHANGE -> 2
            tendency = decoded['rvr']['tend'][i]
            if tendency == 'UPWARD':
                output[f'rvr_{i+1}_tend'] = 0
            elif tendency == 'DOWNWARD':
                output[f'rvr_{i+1}_tend'] = 1
            elif tendency == 'MISSING_VALUE':
                output[f'rvr_{i+1}_tend'] = 3

    num_vrbrvr = 0
    if 'vrbrvr' in decoded:
        num_vrbrvr = len(decoded['vrbrvr']['str'])
        for i in range(num_vrbrvr):
            #   ecCodes 001064: Runway designator
            output[f'vrbrvr_{i+1}_designator'] = decoded['vrbrvr']['rwy'][i]

            #   ecCodes 008014: Qualifier for runway visual range
            #   encoding: 0 - NORMAL, 1 - ABOVE THE UPPER LIMIT, 2 - BELOW THE LOWER LIMIT
            qualifier = decoded['vrbrvr']['oper'][i]
            if qualifier == 'P':
                output[f'vrbrvr_{i+1}_qualifier'] = 1
            elif qualifier == 'M':
                output[f'vrbrvr_{i+1}_qualifier'] = 2
            elif qualifier == None:
                output[f'vrbrvr_{i+1}_qualifier'] = 0

            low_vis = float(decoded['vrbrvr']['lo'][i])
            high_vis = float(decoded['vrbrvr']['hi'][i])
            
            #   ecCodes 020061: Runway Visual Range recorded in meters
            #   if distance is reported in meters, we can copy directly to bufr
            if decoded['vrbrvr']['uom'][i] == 'm':
                output[f'vrbrvr_{i+1}_low'] = int(low_vis)
                output[f'vrbrvr_{i+1}_high'] = int(high_vis)

            #   if distance is reported in ft, have to convert to meters for bufr
            elif decoded['vrbrvr']['uom'][i] == '[ft_i]':
                output[f'vrbrvr_{i+1}_low'] = math.floor(low_vis*0.3048)
                output[f'vrbrvr_{i+1}_high'] = math.floor(high_vis*0.3048)

    #   ecCodes 020019: Significant weather as either precipitation events or observed phenomena
    num_pcp = 0
    if 'pcp' in decoded:
        num_pcp = len(decoded['pcp']['str'])
        for i in range(num_pcp):
            output[f'precip_{i+1}'] = decoded['pcp']['str'][0]
    num_obv = 0
    if 'obv' in decoded:
        num_obv = len(decoded['obv']['str'])
        for i in range(num_obv):
            output[f'observed_phenom_{i+1}'] = decoded['obv']['str'][i]

    num_sky = 0
    if 'sky' in decoded:
        num_sky = len(decoded['sky']['str'])
        for i in range(num_sky):
            #   special cases where sky cover is recorded as SKC or CLR both with their own meanings. Set flags
            #   for these cases in the output and return zero or one sky groups
            if decoded['sky']['str'][0] == 'SKC':
                output['SKC'] = True
                num_sky = 0
                break
            if decoded['sky']['str'][0] == 'CLR':
                output['CLR'] = True
                num_sky = 1
                break

            #   ecCodes 020091: Vercial visibility reported in ft
            if decoded['sky']['str'][0][:2] == 'VV':
                #   METAR reports vertical visibility in 100s of ft, multiply by 100 to get correct bufr value
                output['vertical_visibility'] = int(decoded['sky']['str'][0][2:]) * 100
                #   also need to reset number of sky groups to zero
                num_sky = 0
                break

            #   ecCodes 020011: Cloud Amount
            #   FEW -> 13, SCT -> 11, BKN -> 12, OVC -> 8*, UNDISCERNABLE OR OBSERVATION NOT MADE -> 15
            #   *FMH1 describes overcast clouds as having full 8 okta cover
            cloud_amount = decoded['sky']['str'][i][:3]
            if cloud_amount == 'FEW':
                output[f'cloud_amount_{i+1}'] = 13
            elif cloud_amount == 'SCT':
                output[f'cloud_amount_{i+1}'] = 11
            elif cloud_amount == 'BKN':
                output[f'cloud_amount_{i+1}'] = 12
            elif cloud_amount == 'OVC':
                output[f'cloud_amount_{i+1}'] = 8
            elif cloud_amount == '///':
                output[f'cloud_amount_{i+1}'] = 15

            #   ecCodes 020092: Height of base cloud reported in ft
            #   check for missing value
            if decoded['sky']['str'][i][3:6] != '///':
                output[f'cloud_height_{i+1}'] = int(decoded['sky']['str'][i][3:6]) * 100

            #   ecCodes 020012: Cloud Type
            #   capture significant convective cloud groups, either cumulonimbus (CB) or towering cumulus (TCU) as reported by METAR
            #   CB -> 8, TCU -> 32
            if len(decoded['sky']['str'][i]) > 6:
                cloud_type = decoded['sky']['str'][i][6:]
                if cloud_type == 'CB':
                    output['sig_convec_'+str(i+1)] = 8
                elif cloud_type == 'TCU':
                    output['sig_convec_'+str(i+1)] = 32


    return output, num_rvr, num_vrbrvr, num_pcp, num_obv, num_sky

def transform(data: str, metadata: str, year: int, month: int) -> Iterator[dict]:
    # ===================
    # First parse metadata file
    # ===================
    if isinstance(metadata, str):
        fh = StringIO(metadata)
        reader = csv.reader(fh, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        col_names = next(reader)
        metadata_dict = {}
        tsi_mapping = {}
        for row in reader:
            if len(row) == 0:
                continue
            single_row = dict(zip(col_names, row))
            tsi = single_row['traditional_station_identifier']
            try:
                wsi = single_row['wigos_station_identifier']
                metadata_dict[wsi] = deepcopy(single_row)
                if tsi in tsi_mapping:
                    LOGGER.warning(("Duplicate entries found for station"
                                    f" {tsi} in station list file"))
                tsi_mapping[tsi] = wsi
            except Exception as e:
                LOGGER.error(e)

        fh.close()
        # metadata = metadata_dict[wsi]
    else:
        LOGGER.error("Invalid metadata")
        raise ValueError

    try:
        is_bulletin = False
        messages = extract_metar(data)
        if messages[0][:6] == 'METAR\n':
            LOGGER.debug("File is a METAR bulletin")
            is_bulletin = True
            bulletin_type = 'METAR'
        elif messages[0][:6] == 'SPECI\n':
            LOGGER.debug("File is a SPECI bulletin")
            is_bulletin = True
            bulletin_type = 'SPECI'
        # print(messages)
    except Exception as e:
        LOGGER.error(e)
        return None
     # Count how many conversions were successful using a dictionary
    conversion_success = {}

    for metar in messages[is_bulletin:]:
        # print(metar)
        result = dict()

        mapping = deepcopy(_mapping)
        # print(mapping)

        conversion_success = {}

        try:
            if is_bulletin:
                msg, num_rvr, num_vrbrvr, num_pcp, num_obv, num_sky = parse_metar(f"{bulletin_type} {metar}", year, month)
            else:
                msg, num_rvr, num_vrbrvr, num_pcp, num_obv, num_sky = parse_metar(metar, year, month)
            tsi = msg['tsi']
        except Exception as e:
            LOGGER.error(f"Error parsing METAR report: {metar}. {str(e)}")

        try:
            wsi = tsi_mapping[tsi]
        except Exception:
            conversion_success[tsi] = False
            LOGGER.warning(f"Station {tsi} not found in station file")

            #   Question for Tom on how wigos mapping is resulting in conversion success being set even though station
            #   is not present
            #   suspect stale data in wsi is retained from previous iteration
            continue

        # extract station info from metadata dict
        try:
            wsi_series, wsi_issuer, wsi_issue_number, wsi_local = wsi.split("-")   # noqa
            # get other required metadata
            latitude = metadata_dict[wsi]["latitude"]
            longitude = metadata_dict[wsi]["longitude"]
            station_height = metadata_dict[wsi]["elevation"]
            # add these values to the data dictionary
            msg['_wsi_series'] = wsi_series
            msg['_wsi_issuer'] = wsi_issuer
            msg['_wsi_issue_number'] = wsi_issue_number
            msg['_wsi_local'] = wsi_local
            msg['_latitude'] = latitude
            msg['_longitude'] = longitude
            msg['_station_height'] = station_height
            conversion_success[tsi] = True
        except Exception:
            conversion_success[tsi] = False
            if wsi == "":
                LOGGER.warning(f"Missing station ID for station {tsi}")
            else:
                LOGGER.warning((f"Invalid WSI ({wsi}) found in station file,"
                                " unable to parse"))
            #   Question for Tom on how wigos mapping is resulting in conversion success being set even though station
            #   is not present
            #   suspect stale data in wsi is retained from previous iteration
            continue

        #   Create mapping for wind speed and gust speed depending on units of measurement
        if 'wind_speed_kn' in msg:
            mapping['data'] = update_data_mapping(mapping['data'], {"eccodes_key": "#2#windSpeed", "value": "data:wind_speed_kn"})
        elif 'wind_speed_m/s' in msg:
            mapping['data'] = update_data_mapping(mapping['data'], {"eccodes_key": "#3#windSpeed", "value": "data:wind_speed_m/s"})
        if 'gust_speed_kn' in msg:
            mapping['data'] = update_data_mapping(mapping['data'], {"eccodes_key": "#2#maximumWindGustSpeed", "value": "data:gust_speed_kn"})
        elif 'gust_speed_m/s' in msg:
            mapping['data'] = update_data_mapping(mapping['data'], {"eccodes_key": "#3#maximumWindGustSpeed", "value": "data:gust_speed_m/s"})
        

        #   Update mappings based on the number of runway visual range and variable runway visual range groups
        for idx in range(num_rvr):
            rvr_mappings = [
                {"eccodes_key": f"#{idx+1}#runwayDesignator", "value": f"data:rvr_{idx+1}_designator"},
                {"eccodes_key": f"#{(idx*2)+1}#qualifierForRunwayVisualRange", "value": f"data:rvr_{idx+1}_qualifier"},
                {"eccodes_key": f"#{(idx*2)+1}#runwayVisualRangeRvr", "value": f"data:rvr_{idx+1}_mean"},
                {"eccodes_key": f"#{idx+1}#tendencyOfRunwayVisualRange", "value": f"data:rvr_{idx+1}_tend"}
            ]

            for m in rvr_mappings:
                mapping['data'] = update_data_mapping(mapping['data'], m)

        for idx in range(num_vrbrvr):
            #   if qualifier is 'P' for above the measuring limit, qualifier will go before the 2nd recorded rvr (max)
            if msg[f'vrbrvr_{idx+1}_qualifier'] == 1:
                vrbrvr_mappings = [
                    {"eccodes_key": f"#{idx+num_rvr+1}#runwayDesignator", "value": f"data:vrbrvr_{idx+1}_designator"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+1}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_low"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+2}#qualifierForRunwayVisualRange", "value": f"data:vrbrvr_{idx+1}_qualifier"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+2}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_high"},
                ]
            #   if qualifer is 'M' for below the measuring limit, qualifier will go before the 1st recorded rvr (min)
            elif msg[f'vrbrvr_{idx+1}_qualifier'] == 2:
                vrbrvr_mappings = [
                    {"eccodes_key": f"#{idx+num_rvr+1}#runwayDesignator", "value": f"data:vrbrvr_{idx+1}_designator"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+1}#qualifierForRunwayVisualRange", "value": f"data:vrbrvr_{idx+1}_qualifier"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+1}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_low"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+2}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_high"},
                ]
            #   if qualifier is not specified, don't include it in this part of the mapping
            elif msg[f'vrbrvr_{idx+1}_qualifier'] == 0:
                vrbrvr_mappings = [
                    {"eccodes_key": f"#{idx+num_rvr+1}#runwayDesignator", "value": f"data:vrbrvr_{idx+1}_designator"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+1}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_low"},
                    {"eccodes_key": f"#{idx+(num_rvr*2)+2}#runwayVisualRangeRvr", "value": f"data:vrbrvr_{idx+1}_high"},
                ]
            for m in vrbrvr_mappings:
                mapping['data'] = update_data_mapping(mapping['data'], m)
        
        #   Update mapping based on number of weather groups, both precipitation and observed phenomena
        for idx in range(num_pcp):
            pcp_mappings = {"eccodes_key": f"#{idx+1}#significantWeather", "value": f"data:precip_{idx+1}"}
            mapping['data'] = update_data_mapping(mapping['data'], pcp_mappings)

        for idx in range(num_obv):
            obv_mappings = {"eccodes_key": f"#{idx+num_pcp+1}#significantWeather", "value": f"data:observed_phenom_{idx+1}"}
            mapping['data'] = update_data_mapping(mapping['data'], obv_mappings)


        #   if SKC reported in cloud/sky group, don't add any additional mappings for clouds and update the general weather indicator
        if 'SKC' in msg:
            msg['general_weather_indicator'] = 3

        #   if CLR reported in cloud/sky group, add only mappings for zero cloud cover based on ecCodes 020011: Cloud Amount and Clear
        #   for ecCodes 020012: Cloud Type
        if 'CLR' in msg:
            cloud_mappings = [
                {"eccodes_key": "#1#cloudAmount", "value": "const:0"},
                {"eccodes_key": "#1#cloudType", "value": "const:43"},
            ]
            for m in cloud_mappings:
                mapping['data'] = update_data_mapping(mapping['data'], m)
        
        #   Otherwise, update mapping based on other reported sky/cloud groups
        else:
            for idx in range(num_sky):
                #   if significant convectivce clouds have been reported, mapping needs to include cloud type
                if f'sig_convec_{idx+1}' in msg:
                    cloud_mappings = [
                        {"eccodes_key": f"#{idx+1}#cloudAmount", "value": f"data:cloud_amount_{idx+1}"},
                        {"eccodes_key": f"#{idx+1}#cloudType", "value": f"data:sig_convec_{idx+1}"},
                        {"eccodes_key": f"#{(idx*2)+2}#heightOfBaseOfCloud", "value": f"data:cloud_height_{idx+1}"}
                    ]
                else:
                    cloud_mappings = [
                        {"eccodes_key": f"#{idx+1}#cloudAmount", "value": f"data:cloud_amount_{idx+1}"},
                        {"eccodes_key": f"#{(idx*2)+2}#heightOfBaseOfCloud", "value": f"data:cloud_height_{idx+1}"}
                    ]
                for m in cloud_mappings:
                    mapping['data'] = update_data_mapping(mapping['data'], m)

        #   Update mapping for vertical visibility if it is recorded instead of a cloud group
        if 'vertical_visibility' in msg:
            vvis_mapping = {"eccodes_key": "#2#verticalVisibility", "value": "data:vertical_visibility"}
            mapping['data'] = update_data_mapping(mapping['data'], vvis_mapping)

        # for item in mapping['data']:
        #     print(item)

        #   set number of replications for each group
        nhvis_307046 = 0
        nrvr_307013 = num_rvr+num_vrbrvr
        nsig_307014 = num_pcp+num_obv
        ncld_307047 = num_sky
        nsig_307016 = 0
        nrwy_307017 = 0
        nsea_307049 = 0     #   short
        nrwy_307050_1 = 0   #   short
        nrwy_307050_2 = 0
        nrwy_307050_3 = 0
        ntrend_307048_1 = 0
        ntrend_307048_2 = 0
        ntrend_307048_3 = 0 #   short
        ntrend_307048_4 = 0 #   short
        ntrend_307048_5 = 0
        ntrend_307048_6 = 0

        unexpanded_descriptors = [307051]
        short_delayed_replications = [
            nsea_307049, nrwy_307050_1, ntrend_307048_3, ntrend_307048_4
        ]
        delayed_replications = [
            nhvis_307046,
            nrvr_307013,
            nsig_307014,
            ncld_307047,
            nsig_307016,
            nrwy_307017,
            nrwy_307050_2,
            nrwy_307050_3,
            ntrend_307048_1,
            ntrend_307048_2,
            ntrend_307048_5,
            ntrend_307048_6
        ]
        extended_delayed_replications = []
        table_version = 37

        try:
            # import pdb; pdb.set_trace()

            # create new BUFR msg
            message = BUFRMessage(
                unexpanded_descriptors,
                short_delayed_replications,
                delayed_replications,
                extended_delayed_replications,
                table_version)
        except Exception as e:
            LOGGER.error(e)
            LOGGER.error("Error creating BUFRMessage")
            conversion_success[tsi] = False
            #   Question for Tom on how wigos mapping is resulting in conversion success being set even though station
            #   is not present
            #   suspect stale data in wsi is retained from previous iteration
            continue

        # parse
        if conversion_success[tsi]:
            try:
                # import pdb; pdb.set_trace()
                message.parse(msg, mapping)
            except Exception as e:
                LOGGER.error(e)
                LOGGER.error("Error parsing message")
                conversion_success[tsi] = False

        # Only convert to BUFR if there's no errors so far
        if conversion_success[tsi]:
            try:
                result["bufr4"] = message.as_bufr()  # encode to BUFR
                status = {"code": PASSED}
            except Exception as e:
                LOGGER.error("Error encoding BUFR, null returned")
                LOGGER.error(e)
                result["bufr4"] = None
                status = {
                    "code": FAILED,
                    "message": f"Error encoding, BUFR set to None:\n\t\tError: {e}\n\t\tMessage: {msg}"  # noqa
                }
                conversion_success[tsi] = False

            # now identifier based on WSI and observation date as identifier
            isodate = message.get_datetime().strftime('%Y%m%dT%H%M%S')
            rmk = f"ICAO_{tsi}_{isodate}"

            # now additional metadata elements
            result["_meta"] = {
                "id": rmk,
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                        message.get_element('#1#longitude'),
                        message.get_element('#1#latitude')
                    ]
                },
                "properties": {
                    "md5": message.md5(),
                    "wigos_station_identifier": wsi,
                    "datetime": message.get_datetime(),
                    "originating_centre":
                    message.get_element("bufrHeaderCentre"),
                    "data_category": message.get_element("dataCategory")
                },
                "result": status
            }

        # now yield result back to caller
        yield result

        # Output conversion status to user
        if conversion_success[tsi]:
            LOGGER.info(f"Station {tsi} report converted")
        else:
            LOGGER.info(f"Station {tsi} report failed to convert")

    # calculate number of successful conversions
    conversion_count = sum(tsi for tsi in conversion_success.values())
    # print number of messages converted
    LOGGER.info((f"{conversion_count} / {len(messages)}"
            " reports converted successfully"))


# metar = "SAAG SABM 012200\
# SPECI DNPO 180700Z 00000KT 5000 BR BKN008\
#     24/24 Q1015 BECMG 7000=\
# SPECI DNPO 180700Z 00000KT 5000 BR BKN008 24/24\
#         Q1015 BECMG 7000=\
# SPECI DNPO 180700Z\
#             00000KT 5000 BR BKN008 24/24 Q1015 BECMG 7000="

# with open("/home/alexander.thompson/metar2bufr_pkg/src/metar2bufr/resources/generic_station_list.csv") as fh:
#     station_metadata = fh.read()

# result = transform(metar, station_metadata, 2023, 10)

# for item in result:
#     # print(item)
#     bufr4 = item['bufr4']

# # f = open('bufrtest.bufr','wb')
# # f.write(bufr4)
# # f.close()

# geojson_results = generic_bufr2geojson.transform(bufr4, serialize=False)
# for item in geojson_results:
#     for key in item.keys():
#         print(item[key]['geojson'])
# test_dict, num_rvr, num_vrbrvr, num_pcp, num_obv, num_sky = parse_metar(metar, 2023, 6)
# for key in test_dict:
#     print(key, ': ', test_dict[key])

# unexpanded_descriptors = [307051]
# short_delayed_replications = []
# delayed_replications = []
# extended_delayed_replications = []
# table_version = 37

# print("==ecCode Keys for descriptor 307051 - Metar representation==")
# message = BUFRMessage(
#             unexpanded_descriptors,
#             short_delayed_replications,
#             delayed_replications,
#             extended_delayed_replications,
#             table_version)

# unexpanded_descriptors = [307012]
# print("==ecCode Keys for descriptor 307012==")
# message = BUFRMessage(
#             unexpanded_descriptors,
#             short_delayed_replications,
#             delayed_replications,
#             extended_delayed_replications,
#             table_version)

# unexpanded_descriptors = [307013]
# print("==ecCode Keys for descriptor 307013==")
# message = BUFRMessage(
#             unexpanded_descriptors,
#             short_delayed_replications,
#             delayed_replications,
#             extended_delayed_replications,
#             table_version)


###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################
import logging
from pygeoapi.process.base import BaseProcessor

from csv2bufr import transform

LOGGER = logging.getLogger(__name__)

PROCESS_METADATA = {
    "version": "0.1.0",
    "id": "csv2bufr-transform",
    "title": {"en": "csv2bufr"},
    "description": {"en": "Process to convert CSV data to BUFR"},  # noqa
    "keywords": [],
    "links": [{
        "type": "text/html",
        "rel": "about",
        "title": "information",
        "href": "https://example.org/process",
        "hreflang": "en-US",
    }],
    "inputs": {
        "data": {
            "title": "CSV Data",
            "description": "Input CSV file contents as string.",  # noqa
            "minOccurs": 1,
            "maxOccurs": 1,
            "metadata": None,
            "keywords": [],
        },
        "mapping": {
            "title": "BUFR Mapping",
            "description": "JSON file describing BUFR mapping.",  # noqa
            "schema": {"type": "string"},
            "minOccurs": 1,
            "maxOccurs": 1,
            "metadata": None,
            "keywords": [],
        },
    },
    "outputs": {
        "messages": {
            "title": "BUFR encoded data in base64",  # noqa
            "schema": {"type": "array"},
            "description": "One result per line in input data"  # noqa
        }
    },
    "example": {
        "inputs": {
            "data":"wsi,year,month,day,hour,minute,latitude,longitude,slp,mslp,ppp,a,brmh\\\n0-20000-0-06700,2022,2,10,6,0,46.2475,6.12774,978.3,1029.9,-0.4,8,412.3",  # noqa
            "mapping": '''{
                "inputDelayedDescriptorReplicationFactor": [],
                "inputShortDelayedDescriptorReplicationFactor": [],
                "inputExtendedDelayedDescriptorReplicationFactor": [],
                "wigos_station_identifier": "data:wsi",
                "number_header_rows": 1,
                "column_names_row": 1,
                "QUOTING": "QUOTE_NONE"
                "header": [
                    {"eccodes_key": "edition", "value": "const:4"},
                    {"eccodes_key": "masterTableNumber", "value": "const:0"},
                    {"eccodes_key": "updateSequenceNumber", "value": "const:0"},
                    {"eccodes_key": "dataCategory", "value": "const:0"},
                    {"eccodes_key": "internationalDataSubCategory", "value": "const:6"},
                    {"eccodes_key": "masterTablesVersionNumber", "value": "const:36"},
                    {"eccodes_key": "typicalYear", "value": "data:year"},
                    {"eccodes_key": "typicalMonth", "value": "data:month"},
                    {"eccodes_key": "typicalDay", "value": "data:day"},
                    {"eccodes_key": "typicalHour", "value": "data:hour"},
                    {"eccodes_key": "typicalMinute", "value": "data:minute"},
                    {"eccodes_key": "numberOfSubsets", "value": "const:1"},
                    {"eccodes_key": "observedData", "value": "const:1"},
                    {"eccodes_key": "compressedData", "value": "const:0"},
                    {"eccodes_key": "unexpandedDescriptors", "value": "array: 301150, 301011, 301012, 301021, 7031, 302001"}
                ],
                "data": [
                    {"eccodes_key": "#1#wigosIdentifierSeries", "value": "metadata:wsi_series", "valid_min": "const:0", "valid_max": "const:0"},
                    {"eccodes_key": "#1#wigosIssuerOfIdentifier", "value": "metadata:wsi_issuer", "valid_min": "const:0", "valid_max": "const:65534"},
                    {"eccodes_key": "#1#wigosIssueNumber", "value": "metadata:wsi_issue_number", "valid_min": "const:0", "valid_max": "const:65534"},
                    {"eccodes_key": "#1#wigosLocalIdentifierCharacter", "value": "metadata:wsi_local"},
                    {"eccodes_key": "#1#year", "value": "data:year", "valid_min": "const:2022", "valid_max": "const:2025"},
                    {"eccodes_key": "#1#month", "value": "data:month", "valid_min": "const:1", "valid_max": "const:12"},
                    {"eccodes_key": "#1#day", "value": "data:day", "valid_min": "const:1", "valid_max": "const:31"},
                    {"eccodes_key": "#1#hour", "value": "data:hour", "valid_min": "const:0", "valid_max": "const:23"},
                    {"eccodes_key": "#1#minute", "value": "data:minute", "valid_min": "const:0", "valid_max": "const:59"},
                    {"eccodes_key": "#1#latitude", "value": "data:latitude", "valid_min": "const:-90.0", "valid_max": "const:90.0"},
                    {"eccodes_key": "#1#longitude", "value": "data:longitude", "valid_min": "const:-180.0", "valid_max": "const:180.0"},
                    {"eccodes_key": "#1#heightOfBarometerAboveMeanSeaLevel", "value": "data:brmh", "valid_min": "const:-400.0", "valid_max": "const:12707.0"},
                    {"eccodes_key": "#1#nonCoordinatePressure", "value": "data:slp", "valid_min": "const:0", "valid_max": "const:163820", "scale": "const:2", "offset": "const:0"},
                    {"eccodes_key": "#1#pressureReducedToMeanSeaLevel", "value": "data:mslp", "valid_min": "const:0", "valid_max": "const:163820", "scale": "const:2", "offset": "const:0"},
                    {"eccodes_key": "#1#3HourPressureChange", "value": "data:ppp", "valid_min": "const:-5000", "valid_max": "const:5220", "scale": "const:2", "offset": "const:0"},
                    {"eccodes_key": "#1#characteristicOfPressureTendency", "value": "data:a", "valid_min": "const:0", "valid_max": "const:14"}
                ]
            }'''  # noqa
        },
        "output": {"messages": []}
    },
}


class csv2bufrProcessor(BaseProcessor):
    def __init__(self, processor_def):
        """
        Initialize object
        :param processor_def: provider definition
        :returns: pygeoapi.process.csv2bufr.csv2bufr
        """

        super().__init__(processor_def, PROCESS_METADATA)

    def execute(self, data):
        """
        This method is invoked by pygeoapi when this class is set as a
        `process` type resource in pygeoapi config file.

        :param data: It is the value of `inputs` passed in payload. e.g.
        {
            "inputs": {
                "data": "csv data to encode",
                "mappings": "csv2bufr mapping json file"
            }
        }

        :return: media_type, json
        """

        mimetype = "application/json"
        try:
            bufr_generator = transform(data["data"], data["mapping"])
            # transform returns a generator, we need to iterate over
            # and add to single output object
            bufr = []
            for result in bufr_generator:
                bufr.append(result)

            output = {"messages": bufr}

            LOGGER.error(output)
        except Exception as e:
            LOGGER.exception(e)
            output = {"messages": None}
        return mimetype, output

    def __repr__(self):
        return "<csv2bufr> {}".format(self.name)
